@@ -1,4 +1,4 @@
-/* tftp-hpa: $Id: tftpd.c,v 1.56 2004/01/08 20:48:30 hpa Exp $ */
+/* tftp-hpa: $Id: tftpd.c,v 1.58 2004/09/14 22:38:46 hpa Exp $ */
 
 /* $OpenBSD: tftpd.c,v 1.13 1999/06/23 17:01:36 deraadt Exp $	*/
 
@@ -45,7 +45,7 @@ static const char *copyright UNUSED =
 /*static char sccsid[] = "from: @(#)tftpd.c	5.13 (Berkeley) 2/26/91";*/
 /*static char rcsid[] = "$OpenBSD: tftpd.c,v 1.13 1999/06/23 17:01:36 deraadt Exp $: tftpd.c,v 1.6 1997/02/16 23:49:21 deraadt Exp $";*/
 static const char *rcsid UNUSED =
-"tftp-hpa $Id: tftpd.c,v 1.56 2004/01/08 20:48:30 hpa Exp $";
+"tftp-hpa $Id: tftpd.c,v 1.58 2004/09/14 22:38:46 hpa Exp $";
 #endif /* not lint */
 
 /*
@@ -207,25 +207,37 @@ set_socket_nonblock(int fd, int flag)
 }
 
 /*
- * Receive packet with synchronous timeout
+ * Receive packet with synchronous timeout; timeout is adjusted
+ * to account for time spent waiting.
  */
 static int recv_time(int s, void *rbuf, int len, unsigned int flags,
-		     unsigned long timeout_us)
+		     unsigned long *timeout_us_p)
 {
   fd_set fdset;
-  struct timeval tmv;
+  struct timeval tmv, t0, t1;
   int rv, err;
+  unsigned long timeout_us = *timeout_us_p;
+  unsigned long timeout_left, dt;
+
+  gettimeofday(&t0, NULL);
+  timeout_left = timeout_us;
 
   for ( ; ; ) {
     FD_ZERO(&fdset);
     FD_SET(s, &fdset);
     
-    tmv.tv_sec  = timeout_us / 1000000;
-    tmv.tv_usec = timeout_us % 1000000;
-    
     do {
+      tmv.tv_sec  = timeout_left / 1000000;
+      tmv.tv_usec = timeout_left % 1000000;
+    
       rv = select(s+1, &fdset, NULL, NULL, &tmv);
-    } while ( rv == -1 && errno == EINTR );
+      err = errno;
+
+      gettimeofday(&t1, NULL);
+
+      dt = (t1.tv_sec - t0.tv_sec)*1000000 + (t1.tv_usec - t0.tv_usec);
+      *timeout_us_p = timeout_left = ( dt >= timeout_us ) ? 1 : (timeout_us - dt);
+    } while ( rv == -1 && err == EINTR );
     
     if ( rv == 0 ) {
       timer(0);			/* Should not return */
@@ -236,6 +248,7 @@ static int recv_time(int s, void *rbuf, int len, unsigned int flags,
     rv = recv(s, rbuf, len, flags);
     err = errno;
     set_socket_nonblock(s, 0);
+    
     if ( rv < 0 ) {
       if ( E_WOULD_BLOCK(err) || err == EINTR ) {
 	continue;		/* Once again, with feeling... */
@@ -1167,18 +1180,20 @@ tftp_sendfile(struct formats *pf, struct tftphdr *oap, int oacklen)
   struct tftphdr *ap;		/* ack packet */
   static u_short block = 1;	/* Static to avoid longjmp funnies */
   u_short ap_opcode, ap_block;
+  unsigned long r_timeout;
   int size, n;
   
   if (oap) {
     timeout = rexmtval;
     (void)sigsetjmp(timeoutbuf,1);
   oack:
+    r_timeout = timeout;
     if (send(peer, oap, oacklen, 0) != oacklen) {
       syslog(LOG_WARNING, "tftpd: oack: %m\n");
       goto abort;
     }
     for ( ; ; ) {
-      n = recv_time(peer, ackbuf, sizeof(ackbuf), 0, timeout);
+      n = recv_time(peer, ackbuf, sizeof(ackbuf), 0, &r_timeout);
       if (n < 0) {
 	syslog(LOG_WARNING, "tftpd: read: %m\n");
 	goto abort;
@@ -1213,13 +1228,14 @@ tftp_sendfile(struct formats *pf, struct tftphdr *oap, int oacklen)
     timeout = rexmtval;
     (void) sigsetjmp(timeoutbuf,1);
     
+    r_timeout = timeout;
     if (send(peer, dp, size + 4, 0) != size + 4) {
       syslog(LOG_WARNING, "tftpd: write: %m");
       goto abort;
     }
     read_ahead(file, pf->f_convert);
     for ( ; ; ) {
-      n = recv_time(peer, ackbuf, sizeof (ackbuf), 0, timeout);
+      n = recv_time(peer, ackbuf, sizeof (ackbuf), 0, &r_timeout);
       if (n < 0) {
 	syslog(LOG_WARNING, "tftpd: read(ack): %m");
 	goto abort;
@@ -1273,6 +1289,7 @@ tftp_recvfile(struct formats *pf, struct tftphdr *oap, int oacklen)
   static u_short block = 0;
   static int acksize;
   u_short dp_opcode, dp_block;
+  unsigned long r_timeout;
 
   dp = w_init();
   do {
@@ -1290,13 +1307,14 @@ tftp_recvfile(struct formats *pf, struct tftphdr *oap, int oacklen)
     block++;
     (void) sigsetjmp(timeoutbuf,1);
   send_ack:
+    r_timeout = timeout;
     if (send(peer, ackbuf, acksize, 0) != acksize) {
       syslog(LOG_WARNING, "tftpd: write(ack): %m");
       goto abort;
     }
     write_behind(file, pf->f_convert);
     for ( ; ; ) {
-      n = recv_time(peer, dp, PKTSIZE, 0, timeout);
+      n = recv_time(peer, dp, PKTSIZE, 0, &r_timeout);
       if (n < 0) {		/* really? */
 	syslog(LOG_WARNING, "tftpd: read: %m");
 	goto abort;
@@ -1331,7 +1349,7 @@ tftp_recvfile(struct formats *pf, struct tftphdr *oap, int oacklen)
   (void) send(peer, ackbuf, 4, 0);
   
   timeout_quit = 1;		/* just quit on timeout */
-  n = recv_time(peer, buf, sizeof (buf), 0, timeout); /* normally times out and quits */
+  n = recv_time(peer, buf, sizeof (buf), 0, &timeout); /* normally times out and quits */
   timeout_quit = 0;
 
   if (n >= 4 &&			/* if read some data */
